/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Doug Daniels <rainking@alumni.rice.edu>
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Return values for evaluateNode 
 */
// the node is before the highlighting; exclude it
const BEFORE_RANGE_EXCLUDE = 1; 
// the node is where highlighting begins (but not ends); exclude it
const BEGIN_RANGE_EXCLUDE = 2;
// the node is in the range of highlighting; include it
const IN_RANGE_INCLUDE = 3;
// the node is where the range of highlighting ends; include it
const END_RANGE_INCLUDE = 4;
// the node is where the range of highlighting ends; exclude it
const END_RANGE_EXCLUDE = 5;

/* NOTE:
   need to check whether the currentNode is the endContainer *before*
   we check whether it is the start container, so that we don't have the
   problem of changing currentNode in midstream and perhaps making
   currentNode != hLiteRange.startNode when it otherwise would be.  that
   is, going backward doesn't have the problems that going forward does,
   because splitText preserves the left-side node in the split. */

var gBeforeRange;


/**
 * Test drive the highlighter on
 * the current mozilla selection.
 */
function testHighlight() {
    var seln = window._content.getSelection();

    var hRange = document.createRange();
    hRange.setStart(seln.anchorNode, seln.anchorOffset);
    hRange.setEnd(seln.focusNode, seln.focusOffset);

    var rangeList = highlightRange(hRange);

    // enumerate rangeList
    dump("enumerating rangeList:\n");
    for (var i=0; i < rangeList.length; i++) {
	dump(" rangeList[" + i + "]==" + rangeList[i].toString() + 
	     " with startContainer==" + rangeList[i].startContainer + 
	     ", startOffset==" + rangeList[i].startOffset + 
	     " with endContainer==" + rangeList[i].endContainer + 
	     ", endOffset==" + rangeList[i].endOffset + "\n");
    }

    // try to surroundContents of rangeList with a span
    for (var i = 0; i < rangeList.length; i++) {
	var nodeToInsert;
	// check the context
	var useDiv = false;
	var children = rangeList[i].startContainer.childNodes;
	for (var j = 0; j <= rangeList[i].startOffset; j++) {
	    // only element nodes have display info
	    if (! isCharDataNode(children.item(j))) {
		var display = document.defaultView.getComputedStyle(children.item(j), "").getPropertyValue("display");
		// if display is block, then can use a div
		if (display == "block") {
		    dump("rangeList[" + i + "] -- found display==block for child " + j + ": " + children.item(j).nodeName + "\n");
		    useDiv = true;
		    break;
		}
	    }
	}

	if (useDiv) {
	    nodeToInsert = document.createElementNS("http://www.w3.org/1999/xhtml", "div");
	}
	else {
	    dump("rangeList[" + i + "] -- defaulting to use span\n");
	    nodeToInsert = document.createElementNS("http://www.w3.org/1999/xhtml", "span");
	}
	
	simpleSurroundContents(nodeToInsert, rangeList[i]);
	setBackgroundColor(nodeToInsert);
	//nodeToInsert.style.background = "yellow";
    }
}
			   


/**
 * Calculates a list of range
 * that exactly covers the input range and all
 * have their startContainer == endContainer, and
 * thus may be surrounded by a span or similar tag.
 *
 * @param range nsIDOMRange to cover
 * @return Array of Range
 */
function highlightRange(range) {
    /* ensure that range is valid (not collapsed, has start and end containers, start
       and end offsets) */
    // FIXME -- todo

    gBeforeRange = true;

    var rangeList = [];
    dump("range: " + range.toString() + " start: " + range.startContainer.nodeName + 
	 " end: " + range.endContainer.nodeName + "\n");
    var ancestorContainer = range.commonAncestorContainer;
    // ancestorContainer must be an element to ensure that splitting text nodes will work
    if (ancestorContainer.childNodes.length == 0) {
	ancestorContainer = ancestorContainer.parentNode;
    }
    var topLevelRetVal = evaluateNode(ancestorContainer, range, rangeList);
    if (topLevelRetVal == END_RANGE_INCLUDE) {
	var newRange = document.createRange();
	newRange.selectNode(range.commonAncestorContainer);
	rangeList.push(newRange);
    }
    // do nothing
    else if (topLevelRetVal == END_RANGE_EXCLUDE) {}
    else { 
	throw new Error("highlightRange: got return value " + topLevelRetVal + " in error");
    }
	
    return rangeList;
}

/**
 * Evaluates a Node to determine whether it is completely covered by a range.
 * 
 * @param currentNode DOM Node to evaluate
 * @param range nsIDOMRange of highlighted selection
 * @param rangeList accumulated array of ranges to cover the highlighted selection
 * @return how to handle the node -- one of codes at top of file
 */
function evaluateNode(currentNode, hLiteRange, rangeList) {
    var newRange = null;
    var isStartContainer = false;
    var isEndContainer = false;

    var currentNodeIsCharData = isCharDataNode(currentNode);

    dump("evaluateNode called for node: " + currentNode.nodeName + 
	 " for which charNodeType==" + currentNodeIsCharData + " and rangeList.length==" + rangeList.length + "\n");

    // if the currentNode is the endContainer of the highlighted range
    if (hLiteRange.endContainer == currentNode) {
	isEndContainer = true;

	dump("found endContainer: " + hLiteRange.endContainer.nodeName + "\n");

	// CHARACTER DATA type node
	if (currentNodeIsCharData) {

	    /* if the highlighted range does not extend to the end of the text node, 
	     split the text node at the endOffset to get the portion of the text
	     node that actually *is* covered. afterward, the currentNode will exactly
	     cover the end of the highlighted selection */
	    if (hLiteRange.endOffset < currentNode.length) {
		currentNode.splitText(hLiteRange.endOffset);
	    }	
	    // otherwise, do nothing.  currentNode is already correct.
	    else {}
		
	}
	// non-CHARACTER DATA type node (Element)
	else {
	    // if the endOffset isn't the last child in the node
	    if (hLiteRange.endOffset < currentNode.childNodes.length) {
		newRange = document.createRange();
		// set the start to be before the first child of currentNode
		newRange.setStart(currentNode, 0);
		// set end to be endOffset
		newRange.setEnd(currentNode, hLiteRange.endOffset);
	    }
	    /* if endOffset covers all of currentNode, do nothing */
	    else {}
	}
    }
    
    /* if current node is the startContainer of the highlighted
       range.  It may *also* have been the endContainer above */
    if (hLiteRange.startContainer == currentNode) {
	
	dump("found startContainer: " + hLiteRange.startContainer.nodeName + "\n");

	isStartContainer = true;
	
	// start-node is a CHARACTER DATA type Node
	if (currentNodeIsCharData) {

	    // if the highlighted text does not begin at offset 0
	    if (hLiteRange.startOffset > 0) {
		/* we must now split the currentNode to capture
		  only what has been highlighted.  this will mess
		  up the highlighted range hLiteRange, since it should actually
		  start at the beginning of the new node we're creating
		  through the split.  we will have to update hLiteRange */
		var newStartNode = currentNode.splitText(hLiteRange.startOffset);
		// first, reset the end if it has been split -- this can
		// only happen if the start and end containers are equal
		if (hLiteRange.endContainer == currentNode) {
		    hLiteRange.setEnd(newStartNode, newStartNode.nodeValue.length);
		}
		// reset the start
		hLiteRange.setStart(newStartNode, 0);

		/* return a before range exclude -- the next child processed by the
		   parent will now be the start container */
		return BEFORE_RANGE_EXCLUDE;
	    }
	    /* otherwise, if the highlighted text begins at 0, do nothing,
	       as we may be able to include this node in a higher level 
	       if it is completely covered */
	    else {}
	} // end if CHARACTER DATA type node

	// start-node is non-CHARACTER DATA (Element) type Node
	else {
	    // if highlighted range starts after the first child
	    if (hLiteRange.startOffset > 0) {
		/* if there is a newRange already, it means that
		   we've already set a valid ending point for this range.
		   as such, we only need set the correct beginning point */
		if (newRange) {
		    newRange.setStart(currentNode, hLiteRange.startOffset);
		}
		/* otherwise, no valid ending point yet, so 
		   we must create the range and extend it to the end of
		   the currentNode's children */
		else {
		    newRange = document.createRange();
		    newRange.setStart(currentNode, hLiteRange.startOffset);
		    newRange.setEnd(currentNode, currentNode.childNodes.length);
		}
	    }
	    /* otherwise, if the startOffset is before the
	       first child of currentNode, so we don't need to adjust anything */
	    else {}
	} // end else NON-CHARACTER DATA type NODE

	// set the global variable saying that we're
	// no longer before the highlighted range
	// this must be set here, and not at the top,
	// because there is one case above that splits
	// a text node and is thus still before the range
	gBeforeRange = false;

    } // end if currentNode is the highlighted range startContainer

    // if there's a newRange, add it to the list
    if (newRange) {
	rangeList.push(newRange);
    }

    // if endContainer was found
    if (isEndContainer) {
	if (newRange) {
	    return END_RANGE_EXCLUDE;
	}
	else {
	    return END_RANGE_INCLUDE;
	}
    }
    
    // if ! endContainer && startContainer
    if (isStartContainer) {
	if (newRange) {
	    return BEGIN_RANGE_EXCLUDE;
	}
	else {
	    return IN_RANGE_INCLUDE;
	}
    }
    
    /* otherwise, we are not at a boundary point
       and are inside of the highlighted range */

    // if we're at a TEXT Node 
    if (currentNodeIsCharData) {
	// if we're still before the range begins, return such
	if (gBeforeRange) {
	    return BEFORE_RANGE_EXCLUDE;
	}
	// otherwise, include us
	else {
	    return IN_RANGE_INCLUDE;
	}
    }
    
    /* otherwise, we're at a non-Tex node. 
       need to visit the children in order */

    /*  offset to begin including in the range -- 0 is before first child*/
    var includeBeginOffset = 0;
    /* offset to finish including in the range -- length is after last child*/
    var includeEndOffset = -1;
    // whether we see the end of the range in one of our children
    var endRangeFound = false;

    for (var i = 0; i < currentNode.childNodes.length; i++) {
	dump("calling evaluateNode for child[" + i + "] of " + currentNode.nodeName + "\n");
	var retVal = evaluateNode(currentNode.childNodes.item(i), hLiteRange, rangeList);
	dump("child[" + i + "] of " + currentNode.nodeName + " returned " + retValStringFromInt(retVal) + "\n");
	if (retVal == BEFORE_RANGE_EXCLUDE) {
	    includeBeginOffset += 1;
	    if (includeBeginOffset != (i + 1)) {
		throw new Error("invalid logic--includeBeginOffset was not == i+1, meaning " +
				"that a beforeRangeExclude was thrown incorrectly at time " + i);
	    }
	}
	else if (retVal == BEGIN_RANGE_EXCLUDE) {
	    includeBeginOffset += 1;
	    if (includeBeginOffset != (i + 1)) {
		throw new Error("invalid logic--includeBeginOffset was not == i+1, meaning " +
				"that a beginRangeExclude was thrown incorrectly at time " + i);
	    }
	}
	else if (retVal ==  IN_RANGE_INCLUDE) { }
	else if (retVal == END_RANGE_INCLUDE) {
	    includeEndOffset = i + 1;
	    // end the for loop 
	    endRangeFound = true;
	    break;
	}
	else if (retVal == END_RANGE_EXCLUDE) {
	    includeEndOffset = i;
	    // end the for loop 
	    endRangeFound = true;
	    break;
	}
	else {
	    throw new Error("Unexpected case: " + retVal + " in evaluation of child[ " + i + "] == " + 
			    currentNode.childNodes.item(i).nodeName);
	}
    } // end for loop over children
    
    /* if includeEndOffset is still -1, it means
       we didn't find the end offset anywhere within
       our descendants, and thus should include to the last one.
       we have to set includeEndOffset correctly here, because
       in the process of visiting all children, a split text node
       may cause the length to increase */
    if (includeEndOffset == -1) {
	includeEndOffset = currentNode.childNodes.length;
    }

    // if no children are included, 
    // we are either before the range or at the end
    if (includeBeginOffset == includeEndOffset) {
	// at then end of the range, not including
	// any children fully
	if (endRangeFound) { 
	    return END_RANGE_EXCLUDE;
	}
	// before the range completely
	else {
	    return BEFORE_RANGE_EXCLUDE;
	}
    }

    // if only some children included
    if (includeBeginOffset != 0 ||
	includeEndOffset != currentNode.childNodes.length) {
	// put these children in a range
	newRange = document.createRange();
	newRange.setStart(currentNode, includeBeginOffset);
	newRange.setEnd(currentNode, includeEndOffset);

	// add the range
	rangeList.push(newRange);

	if(endRangeFound) {
	    return END_RANGE_EXCLUDE;
	}
	else {
	    return BEGIN_RANGE_EXCLUDE;
	}
    }

    // otherwise, all children included
    if (endRangeFound) {
	return END_RANGE_INCLUDE;
    }
    else {
	return IN_RANGE_INCLUDE;
    }
}


/**
 * Is this node a character-data or text type?
 */
function isCharDataNode(node) {
    return ((node.nodeType == node.TEXT_NODE) ||
	    (node.nodeType == node.CDATA_SECTION_NODE));
}

/**
 * A version of the Range method surround contents
 * that works in the limited context we need it.
 * Assumes that the range has startContainer == endContainer
 * (which ours always do) and has non-Text node startContainer
 * and endContainer.
 * Surrounds such ranges with the input tag.
 *
 * @param tag nsIDOMNode to surround range with
 * @param range nsIDOMRange to be surrounded
 */
function simpleSurroundContents(tag, range) {
    // if the startContainer or endContainer are not
    // the same element, scream
    if (range.startContainer != range.endContainer) {
	throw new Error("startContainer and endContainer not equal");
    }
    if ( (range.startContainer.nodeType == range.startContainer.TEXT_NODE) ||
	 (range.endContainer.nodeType == range.endContainer.TEXT_NODE) ) {
	throw new Error("startContainer and/or endContainer are text nodes");
    }

    
    var parent = range.startContainer;

    // insert the tag before the start of the range
    var nodesToMove = [];
    for (var i = range.startOffset; i < range.endOffset; i++) {
	var node = range.startContainer.childNodes.item(i);
	nodesToMove.push(node);
    }
    
    parent.insertBefore(tag, range.startContainer.childNodes.item(range.startOffset));

    for (var i = 0; i < nodesToMove.length; i++) {
	tag.appendChild(parent.removeChild(nodesToMove[i]));
    }
}
	
/**
 * A utility function to turn the
 * return values from evaluateNode
 * into human-readable strings.
 *
 * @param code evaluateNode return value
 * @return string version
 */
function retValStringFromInt(code) {
    switch (code) {
	case BEFORE_RANGE_EXCLUDE:
	    return "BEFORE_RANGE_EXCLUDE";
case BEGIN_RANGE_EXCLUDE:
    return "BEGIN_RANGE_EXCLUDE";
case IN_RANGE_INCLUDE:
    return "IN_RANGE_INCLUDE";
case END_RANGE_INCLUDE:
    return "END_RANGE_INCLUDE";
case END_RANGE_EXCLUDE:
    return "END_RANGE_EXCLUDE";
    default:
	throw new Error("Invalid return value");
    }
}

/**
 * Utility function to set the background color
 * for an element and all of its descendants.
 *
 * @param element nsIDOMElement
 */
function setBackgroundColor(element) {
    if (element.style) {
	    element.style.background = "yellow";
    }

    for (var i = 0; i < element.childNodes.length; i++) {
	setBackgroundColor(element.childNodes.item(i));
    }
}


    
	
